import getpass
import logging
import sys
import webbrowser

import httpx
import keyring
from keyrings.alt.file import PlaintextKeyring

from .log import setup_logger, teardown_logger
from .sets import Settings, get_console
from .util import ANSI, print_url

tlogger = logging.getLogger("auth")
tag = "Auth"


def login(settings=None, retry=False):
    settings = settings or Settings()
    setup_logger(settings=settings, logger=tlogger)
    try:
        auth = keyring.get_password(f"{settings.tag}", f"{settings.tag}")
    except keyring.errors.NoKeyringError:  # fallback
        keyring.set_keyring(PlaintextKeyring())
        auth = keyring.get_password(f"{settings.tag}", f"{settings.tag}")
    if settings.auth is None:
        if auth == "":
            keyring.delete_password(f"{settings.tag}", f"{settings.tag}")
        elif auth is not None:
            settings.auth = auth
    if settings.auth == "":
        tlogger.critical(
            "%s: authentication failed: the provided token cannot be empty", tag
        )
        settings.auth = "_key"
    client = httpx.Client(
        verify=True if not settings.insecure_disable_ssl else False,
        proxy=settings.http_proxy or settings.https_proxy or None
    )
    r = client.post(
        url=settings.url_login,
        headers={
            "Authorization": f"Bearer {settings.auth}",
        },
    )
    try:
        tlogger.info(f"{tag}: logged in as {r.json()['organization']['slug']}")
        keyring.set_password(f"{settings.tag}", f"{settings.tag}", f"{settings.auth}")
        teardown_logger(tlogger)
    except Exception as e:
        if retry:
            tlogger.warning("%s: authentication failed", tag)
        hint1 = f"{ANSI.cyan}- Please copy the API key provided in the web portal and paste it below"
        hint2 = f"- You can alternatively manually open {print_url(settings.url_token)}"
        hint3 = f"{ANSI.green}- You may exit at any time by pressing CTRL+C / ⌃+C"
        tlogger.info(
            f"{tag}: initializing authentication\n\n {hint1}\n\n {hint2}\n\n {hint3}\n"
        )
        webbrowser.open(url=settings.url_token)
        if get_console() == "jupyter":
            settings.auth = getpass.getpass(prompt="Enter API key: ")
        else:
            settings.auth = input(f"{ANSI.yellow}Enter API key: ")
        try:
            keyring.set_password(
                f"{settings.tag}", f"{settings.tag}", f"{settings.auth}"
            )
        except Exception as e:
            tlogger.critical(
                "%s: failed to save key to system keyring service: %s", tag, e
            )
        teardown_logger(tlogger)
        login(retry=True)


def logout(settings=None):
    settings = settings or Settings()
    setup_logger(settings=settings, logger=tlogger)
    try:
        keyring.delete_password(f"{settings.tag}", f"{settings.tag}")
    except keyring.errors.NoKeyringError:
        keyring.set_keyring(PlaintextKeyring())
        keyring.delete_password(f"{settings.tag}", f"{settings.tag}")
    except Exception as e:
        tlogger.warning("%s: failed to delete key from system keyring service: %s", tag, e)
    tlogger.info(f"{tag}: logged out")
    teardown_logger(tlogger)



import logging
import queue
import sqlite3
import threading
import time

from .sets import Settings

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "Store"


class DataStore:
    def __init__(self, config: dict, settings: Settings) -> None:
        self.settings = settings

        self.db = f"{settings.work_dir()}/{settings.store_db}"

        self.conn = sqlite3.connect(
            self.db, check_same_thread=False
        )  # isolation_level=None
        self.conn.execute("PRAGMA journal_mode=WAL;")
        self.cursor = self.conn.cursor()

        self._stop_event = threading.Event()
        self._lock = threading.Lock()
        self._queue = queue.Queue()
        self._thread = None
        self.start()

    def start(self):
        if self._thread is None:
            self._thread = threading.Thread(target=self._worker, daemon=True)
            self._thread.start()

        self.cursor.execute(f"""
            CREATE TABLE IF NOT EXISTS {self.settings.store_table_data}(
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                time INTEGER NOT NULL,
                step INTEGER NOT NULL,
                key TEXT NOT NULL,
                value REAL NOT NULL
            );
        """)
        self.cursor.execute(f"""
            CREATE TABLE IF NOT EXISTS {self.settings.store_table_file}(
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                time INTEGER NOT NULL,
                step INTEGER NOT NULL,
                name TEXT NOT NULL,
                aid REAL NOT NULL
            );
        """)
        self.conn.commit()

    def insert(self, data=None, file=None, timestamp=None, step=None):
        self._queue.put((data, file, timestamp, step))

    def stop(self):
        while not self._queue.empty():
            pass
        self._stop_event.set()
        if self._thread is not None:
            self._thread.join(timeout=None)  # TODO: investigate hanging
            self._thread = None
        self.conn.commit()
        self.conn.close()
        logger.info(f"{tag}: find saved database at {self.db}")

    def _worker(self):
        while not self._stop_event.is_set():
            batch_data, batch_file = [], []
            start = time.time()
            while (
                time.time() - start < self.settings.store_aggregate_interval
                and len(batch_data) < self.settings.store_max_size
                and len(batch_file) < self.settings.store_max_size
            ):
                try:
                    d, f, t, s = self._queue.get(
                        timeout=max(
                            0,
                            self.settings.store_aggregate_interval
                            - (time.time() - start),
                        )
                    )
                    if d != {}:
                        batch_data.append(
                            {
                                "t": t,
                                "s": s,
                                "d": d,
                            }
                        )
                    if f != {}:
                        batch_file.append(
                            {
                                "t": t,
                                "s": s,
                                "f": f,
                            }
                        )
                except queue.Empty:
                    continue
            self._insert(batch_data, batch_file)

    def _insert(self, d, f):
        with self._lock:
            self.conn.execute("BEGIN")
            try:
                if d != []:
                    self.cursor.executemany(
                        f"""
                        INSERT INTO {self.settings.store_table_data} (time, step, key, value) VALUES (?, ?, ?, ?)
                        """,
                        [(e["t"], e["s"], k, v) for e in d for k, v in e["d"].items()],
                    )
                    logger.info(f"{tag}: inserted {len(d)} line(s)")
                if f != []:
                    self.cursor.executemany(
                        f"""
                        INSERT INTO {self.settings.store_table_file} (time, step, name, aid) VALUES (?, ?, ?, ?)
                        """,
                        [
                            (e["t"], e["s"], f"{fe._name}{fe._ext}", fe._id)
                            for e in f
                            # for fe in e["f"].values()
                            for fel in e["f"].values()
                            for fe in fel
                        ],
                    )
                    logger.info(f"{tag}: inserted {len(f)} file(s)")
                self.conn.commit()
            except Exception as e:
                self.conn.rollback()
                logger.error("%s: failed to insert batch: %s", tag, e)



import logging
import os
from datetime import datetime

from . import sets
from .ops import Ops
from .sets import Settings
from .util import gen_id

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "Init"


class OpsInit:
    def __init__(self, config) -> None:
        self.kwargs = None
        self.config: dict[str, any] = config

    def init(self) -> Ops:
        op = Ops(config=self.config, settings=self.settings)
        op.start()
        return op

    def setup(self, settings) -> None:
        init_settings = Settings()
        setup_settings = sets.setup(settings=init_settings).settings

        # TODO: handle login and settings validation here
        setup_settings.update(settings)
        self.settings = setup_settings
        self.settings.meta = []  # TODO: find a better way to de-reference meta


def init(
    dir: str | None = None,
    project: str | None = None,
    name: str | None = None,
    id: str | None = None,
    config: dict | str | None = None,
    settings: Settings | dict[str, any] | None = {},
) -> Ops:
    if not isinstance(settings, Settings):  # isinstance(settings, dict)
        default = Settings()
        default.update(settings)
        settings = default

    settings.dir = dir if dir else settings.dir
    settings.project = project if project else settings.project

    settings._op_name = name if name else datetime.now().strftime("%Y%m%d")
    # settings._op_id = id if id else gen_id(seed=settings.project)

    try:
        op = OpsInit(config=config)
        op.setup(settings=settings)
        return op.init()
    except Exception as e:
        logger.critical("%s: failed, %s", tag, e)  # add early logger
        raise e



import hashlib
import logging
import mimetypes
import os
import re
import shutil
import tempfile
import uuid

from PIL import Image as PILImage

from .util import get_class

logger = logging.getLogger(f"{__name__.split('.')[0]}")

VALID_CHAR = re.compile(r"^[a-zA-Z0-9_\-.]+$")
INVALID_CHAR = re.compile(r"[^a-zA-Z0-9_\-.]")


class File:
    def __init__(
        self,
        path: str,
        name: str | None = None,
    ) -> None:
        self._path = os.path.abspath(path)
        self._id = self._hash()

        if not name:
            name = self._id
        elif not VALID_CHAR.match(name):
            e = ValueError(
                f"invalid file name: {name}; file name may only contain alphanumeric characters, dashes, underscores, and periods; proceeding with sanitized name"
            )
            logger.warning("File: %s", e)
            name = INVALID_CHAR.sub("-", name)
        self._name = name
        self._ext = os.path.splitext(self._path)[-1]
        self._type = self._mimetype()
        self._stat = os.stat(self._path)
        self._size = self._stat.st_size  # os.path.getsize(self._path)
        self._url = None

    def _mimetype(self) -> str:
        return mimetypes.guess_type(self._path)[0] or "application/octet-stream"

    def _hash(self) -> str:  # do not truncate
        with open(self._path, "rb") as f:
            return hashlib.sha256(f.read()).hexdigest()

    def _mkcopy(self, work_dir) -> None:
        if not hasattr(self, "_tmp"):
            self._tmp = f"{work_dir}/files/{self._name}-{self._id}{self._ext}"
            shutil.copyfile(self._path, self._tmp)
            if hasattr(self, "_image"):
                os.remove(self._path)
            self._path = os.path.abspath(self._tmp)


class Image(File):
    def __init__(
        self,
        data: any,  # Union[PILImage.Image, np.ndarray],
        caption: str | None = None,
    ) -> None:
        self._name = caption or "image"
        self._id = f"{uuid.uuid4()}{uuid.uuid4()}".replace("-", "")
        self._ext = ".png"

        if isinstance(data, str):
            logger.debug("Image: used file")
            self._image = "file"  # self._image = PILImage.open(data)
            self._path = os.path.abspath(data)
        else:
            self._path = None
            if isinstance(data, PILImage.Image):
                logger.debug("Image: used PILImage")
                self._image = data
            else:
                class_name = get_class(data)
                if class_name.startswith("matplotlib."):
                    logger.debug("Image: attempted conversion from matplotlib")
                    self._image = make_compat_matplotlib(data)
                elif class_name.startswith("torch.") and (
                    "Tensor" in class_name or "Variable" in class_name
                ):
                    logger.debug("Image: attempted conversion from torch")
                    self._image = make_compat_torch(data)
                else:
                    logger.debug("Image: attempted conversion from array")
                    self._image = make_compat_numpy(data)

    def load(self, work_dir=None):
        if not self._path:
            if work_dir:
                self._tmp = f"{work_dir}/files/{self._name}-{self._id}{self._ext}"
                self._image.save(self._tmp, format=self._ext[1:])
                self._path = os.path.abspath(self._tmp)
            else:
                with tempfile.NamedTemporaryFile(delete=False, suffix=self._ext) as tmp:
                    self._image.save(tmp.name, format=self._ext[1:])
                    self._path = os.path.abspath(tmp.name)

        super().__init__(path=self._path, name=self._name)
        if not self._type.startswith("image/"):
            logger.error(
                f"Image: proceeding with potentially incompatible mime type: {self._type}"
            )


def make_compat_matplotlib(val: any) -> any:
    # from matplotlib.spines import Spine # only required for is_frame_like workaround
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure

    if val == plt:
        val = val.gcf()
    elif not isinstance(val, Figure):
        if hasattr(val, "figure"):
            val = val.figure
            if not isinstance(val, Figure):
                e = ValueError(
                    "Invalid matplotlib object; must be a matplotlib.pyplot or matplotlib.figure.Figure object"
                )
                logger.critical("Image failed: %s", e)
                raise e

    from io import BytesIO

    buf = BytesIO()
    val.savefig(buf, format="png")
    image = PILImage.open(buf, formats=["PNG"])
    return image


def make_compat_torch(val: any) -> any:
    from torchvision.utils import make_grid

    if hasattr(val, "requires_grad") and val.requires_grad:
        val = val.detach()
    if hasattr(val, "dtype") and str(val.dtype).startswith("torch.uint8"):
        val = val.to(float)
    data = make_grid(val, normalize=True)
    image = PILImage.fromarray(
        data.mul(255).clamp(0, 255).byte().permute(1, 2, 0).cpu().numpy()
    )
    return image


def make_compat_numpy(val: any) -> any:
    import numpy as np

    if hasattr(val, "numpy"):
        val = val.numpy()
    if val.ndim > 2:
        val = val.squeeze()

    if np.min(val) < 0:
        val = (val - np.min(val)) / (np.max(val) - np.min(val))
    if np.max(val) <= 1:
        val = (val * 255).astype(np.int32)
    val.clip(0, 255).astype(np.uint8)

    image = PILImage.fromarray(val, mode="RGBA" if val.shape[-1] == 4 else "RGB")
    return image



import json
import logging
import os
import random
import string
import time
import uuid
from typing import Sequence, Union

from .sets import get_console

logger = logging.getLogger(f"{__name__.split('.')[0]}")
tag = "Util"


class ANSI:
    base = "\033["  # "\x1b["
    reset = f"{base}0m"
    bold = f"{base}1m"
    faint = f"{base}2m"
    italic = f"{base}3m"
    underline = f"{base}4m"
    slow_blink = f"{base}5m"
    rapid_blink = f"{base}6m"

    black = f"{base}30m"
    red = f"{base}31m"
    green = f"{base}32m"
    yellow = f"{base}33m"
    blue = f"{base}34m"
    purple = f"{base}35m"
    cyan = f"{base}36m"
    white = f"{base}37m"

    if not __import__("sys").stdout.isatty() and get_console() == "python":
        for _ in dir():
            if isinstance(_, str) and _[0] != "_":
                locals()[_] = ""
    else:
        if __import__("platform").system() == "Windows":
            os.system("")


def print_url(url):
    return f"{ANSI.underline}{url}{ANSI.reset}"


def gen_id(seed=None, length=8) -> str:
    random.seed(uuid.uuid4().hex) if seed is None else random.seed(seed + uuid.uuid4().hex)
    base = string.ascii_lowercase + string.ascii_uppercase + string.digits
    return "".join(random.choice(base) for _ in range(length))


def gen_ulid(base="0123456789ABCDEFGHJKMNPQRSTVWXYZ") -> str:  # py-ulid
    ulid = (int(time.time() * 1000) << 80) | random.getrandbits(80)

    encoded = []
    while ulid > 0:
        ulid, remainder = divmod(ulid, 32)
        encoded.append(base[remainder])
    return "".join(encoded[::-1]).rjust(26, base[0])


def to_dict(obj):
    attrs = {}
    for name in dir(obj):
        if not name.startswith("__") and not callable(getattr(obj, name)):
            attrs[name] = getattr(obj, name)
    return attrs


def to_json(data, file):
    if os.path.exists(file):
        with open(file, "r+") as f:
            try:
                content = json.load(f)
                if not isinstance(content, list):
                    logger.error(f"{tag}: file content must be a json list")
                    return
            except json.JSONDecodeError:
                logger.error(f"{tag}: file is not in json format")
                return
            content.extend(data)
            f.seek(0)
            json.dump(content, f, indent=4)
            f.truncate()
    else:
        with open(file, "w") as f:
            json.dump(data, f, indent=4)


def to_human(n):
    symbols = ("K", "M", "G", "T", "P", "E", "Z", "Y")
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if abs(n) >= prefix[s]:
            value = float(n) / prefix[s]
            return "%.1f%s" % (value, s)
    return "%sB" % n


def dict_to_json(data: dict[str, any]) -> dict:
    for key in list(data):  # avoid RuntimeError if dict size changes
        val = data[key]
        if isinstance(val, dict):
            data[key] = dict_to_json(val)
        else:
            data[key] = val_to_json(val)
    return data


def val_to_json(val: any) -> Union[Sequence, dict]:
    if isinstance(val, (int, float, str, bool)):
        return val
    elif isinstance(val, (list, tuple, range)):
        raise NotImplementedError()  # TODO: for files


def get_class(val: any) -> str:
    module_class = val.__class__.__module__ + "." + val.__class__.__name__
    return (
        val.__name__
        if module_class in ["builtins.module", "__builtin__.module"]
        else module_class
    )
